/* Copyright 2000 - Mathias Broxvall */

/*
TODO: ANOTHER AI VALUE = ENEMY_STRENGTH PROPAGATED THROUGH ENEMY PIPES??
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "ai.h"
#include "constant.h"
#include <math.h>
/**** x include files ****/
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#if USE_LONGJMP
#include <setjmp.h>
#endif

#include "extern.h"

/**********************************************/
/*
 * P.Bureau - Generic code to handle mutiple algo
 * For new algorithm, declare init function and algo
 * function in ai.h, increase AI_MAX_ALGO and modify
 * both table below with new functions.
 */

/* empty init and algo for human players */
int no_init(int side,float skill)
{
    
}
void no_algo(int side)
{
    
}

/* table fo init function */
int (*ai_init_table[AI_MAX_ALGO+1])(int,float)={
    no_init,
    init_algo1,
    init_algo2
};

/* table for algo functions */
void (*ai_algo_table[AI_MAX_ALGO+1])(int)={
    no_algo,
    ai_algo1,
    ai_algo2
};


/***********************************************/

/* Some extra protos */
extern void run_shoot (cell_type *,int ply,int x,int y,int calc,int is_artillery);

/******************/
/*  AI constants  */
/******************/

#define TROOP_VALUE 0
#define CELL_VALUE 1
#define THREAT 2
#define NUM_AI_VALUES 3

#define MOVEMENT_NONE          0
#define MOVEMENT_NORMAL_ON     1
#define MOVEMENT_NORMAL_OFF    2
#define MOVEMENT_CHARGE        3
#define MOVEMENT_ABORT_ATTACK  4
#define MOVEMENT_GUNNERY       5

#define DECAY_NORMAL                0
#define DECAY_OWN_STRENGTH_INVERSE  1

/*****************************************************************
   Explanation of skillLevel:
   Zero for every human player otherwise computer controlled.
   Skill level is first scaled by a constant (to make the scale
   nicer for beginners), if the scaled skill level is less than
   one it represents the likelyhood every "turn" that AI will
   be allowed to move. Ie, scaled skill level 0.5 means 50% chance
   to move every turn. If skill level is greater than one the AI 
   can make as many moves as the interger part and optionally one
   more with the fractional part as likelyhood. Ie, skill level 1.5 
   means 50% chance to make two moves, otherwise only one move.
******************************************************************/

#define MAX_GUNNERY_TARGETS 32
typedef struct s_AIData {  
  float skillLevel;

  /* These are the new gunnery targets calculated this turn */
  int new_num_gunnery_targets,new_gunnery_targets[MAX_GUNNERY_TARGETS][3];

  /* These are the targets provided earlier turns */
  int num_gunnery_targets,gunnery_targets[MAX_GUNNERY_TARGETS][3];
} AIData;

AIData aiData[MAX_PLAYERS];
int currentSide;

extern set_move_on(cell_type*,int*,int);
extern set_move_off(cell_type*,int*,int);

/**********************************/
/* Propagation of positive values */
/**********************************/

void propagate_positive_ai_value(int flags,cell_type *cell,int valueKind,
				 double value,double fallof,int steps) 
{
  double tmp_value = cell->ai_values[valueKind] + value;
  int i;

  if(cell->level < 0) return; /* Ignore water cells */
  if(cell->ai_visited == 0 || tmp_value > cell->ai_tmp_values[valueKind]) {
    cell->ai_tmp_values[valueKind] = tmp_value;
    cell->ai_visited = 1;
    value -= fallof;
    if(flags & DECAY_OWN_STRENGTH_INVERSE) {
      value *= 1.0 - cell->value[currentSide] / (double) Config->max_value[currentSide];
      fallof *= 1.0 - cell->value[currentSide] / (double) Config->max_value[currentSide];
    }
    if(steps > 0)
      for(i=0;i<Config->direction_count;i++)
	if(cell->connect[i] && cell->connect[i] != cell)
	  propagate_positive_ai_value(flags,cell->connect[i],valueKind,value,fallof,steps-1);
  }
}

/**********************************/
/* Propagation of negative values */
/**********************************/

void propagate_negative_ai_value(int flags,cell_type *cell,int valueKind,
				 double value,double fallof,int steps) 
{
  double tmp_value = cell->ai_values[valueKind] + value;
  int i;

  if(cell->level < 0) return; /* Ignore water cells */
  if(cell->ai_visited == 0 || tmp_value < cell->ai_tmp_values[valueKind]) {
    cell->ai_tmp_values[valueKind] = tmp_value;
    cell->ai_visited = 1;
    value -= fallof;
    if(flags & DECAY_OWN_STRENGTH_INVERSE) {
      value *= 1.0 - cell->value[currentSide] / (double) Config->max_value[currentSide];
      fallof *= 1.0 - cell->value[currentSide] / (double) Config->max_value[currentSide];
    }
    if(steps > 0)
      for(i=0;i<Config->direction_count;i++)
	if(cell->connect[i] && cell->connect[i] != cell)
	  propagate_negative_ai_value(flags,cell->connect[i],valueKind,value,fallof,steps-1);
  }
}

/***********************/
/*     Cleaning up     */
/***********************/

void propagate_cleanup(cell_type *cell,int valueKind,int steps) {
  int i;

  if(cell->level < 0) return;
  if(cell->ai_visited) {
    cell->ai_visited = 0;
    if(valueKind == -1)
      cell->ai_dist = 0;
    else
      cell->ai_values[valueKind] = cell->ai_tmp_values[valueKind];
    if(steps > 0)
      for(i=0;i<Config->direction_count;i++)
	if(cell->connect[i] && cell->connect[i] != cell)
	  propagate_cleanup(cell->connect[i],valueKind,steps-1);
  }
}

/*******************************************/
/* Main entrypoint of propagation routines */
/*******************************************/

void propagate_ai_value(int flags,cell_type *cell,int valueKind,double value,int radius) {

    if(value > 0)
        propagate_positive_ai_value(flags,cell,valueKind,value,value/radius,radius);
    else
        propagate_negative_ai_value(flags,cell,valueKind,value,value/radius,radius);

    propagate_cleanup(cell,valueKind,radius);
}

/********************************************************/
/* graph_distance: Finds the distance between two cells */
/********************************************************/

void graph_distance1(cell_type *source,int steps) {
  int i,nd;
  cell_type *cell;

  nd = source->ai_dist + 1;
  source->ai_visited = 1;
  if(steps == 0) return;

  for(i=0;i<Config->direction_count;i++) {
    cell = source->connect[i];
    if(cell == NULL || cell == source) continue;
    if(cell->ai_dist == 0 || nd < cell->ai_dist) {
      cell->ai_dist = nd;
      graph_distance1(cell,steps-1);
    }
  }
}

int graph_distance(cell_type *source,cell_type *dest,int max_range) {
  int d;
  source->ai_dist = 1;
  source->ai_visited = 1;
  graph_distance1(source,max_range+1);
  d = dest->ai_dist - 1;
  propagate_cleanup(source,-1,max_range);
  return d;
}


/********************************/
/* Init function for ai algo 1  */
/********************************/

int init_algo1(int side,float skill)
{
    int width,height,x,y,i;
    cell_type *cell;

    aiData[side].skillLevel = skill * 1.0;
    width = Config->board_x_size;
    height = Config->board_y_size;

    for(x=0;x<width;x++)
        for(y=0;y<height;y++) {
            cell = Board->cells[y][x];
            if(!cell) continue;
            for(i=0;i<NUM_AI_VALUES;i++)
                cell->ai_feedback_values[side][i] = 0.0;
        }

    aiData[side].new_num_gunnery_targets = 0;

    return 0;
}

/*******************************/
/*    The main AI function     */
/*******************************/

int tick=0;

void ai_algo1(int side) {

  int x,y,i,j,k,outgoing,range,X,Y;
  int strength,enemy_strength,owner,this_strength,total_strength;
  cell_type *cell,*cell2,*cell3;

  int width,height;
  int dir[MAX_DIRECTIONS];
  int num_actions;

  double best_preference,this_preference,total_preference,threat,f1;
  cell_type *source_cell, *target_cell;
  int direction;
  int movement_kind;
  char flag1,flag2;

  if(aiData[side].skillLevel == 0) return;
  num_actions=(int)aiData[side].skillLevel;
  if(aiData[side].skillLevel - num_actions > get_random(100)/100.0) num_actions++;
  if(num_actions == 0) return;

  currentSide = side;
  width = Config->board_x_size;
  height = Config->board_y_size;
  tick++;

  
  
  /*  printf("Tick: %d\n",tick); */
  /*for(y=0;y<height;y++) {
    for(x=0;x<width;x++) {
      cell = Board->cells[y][x];
      printf("  %2.2f",cell->ai_feedback_values[side][TROOP_VALUE]);
    }
    printf("\n");
    }*/
  
  /* Distinguish between new and old gunnery targets */  
  aiData[side].num_gunnery_targets = aiData[side].new_num_gunnery_targets;
  aiData[side].new_num_gunnery_targets = 0;
  for(i=0;i<MAX_GUNNERY_TARGETS;i++)
    for(j=0;j<3;j++)
      aiData[side].gunnery_targets[i][j] = aiData[side].new_gunnery_targets[i][j];
  
  for(x=0;x<width;x++)
    for(y=0;y<height;y++) {
      cell = Board->cells[y][x];
      cell->ai_visited = 0;
      cell->ai_already_moved = 0;
      cell->ai_dist = 0;
      for(i=0;i<NUM_AI_VALUES;i++) {
	cell->ai_values[i] = cell->ai_feedback_values[side][i];
	cell->ai_feedback_values[side][i] = cell->ai_tmp_values[i] = 0.0;
      }
    }

  /* Propagate feedback values through existing pipes one step... */
  /*  for(x=0;x<width;x++)
    for(y=0;y<height;y++) {
      cell = Board->cells[y][x];
      if(cell->level < 0) continue;
      if(cell->side != side) continue;
      for(k=0;k<Config->direction_count;k++)
	if(cell->connect[k] == cell ||
	   cell->connect[k] == NULL) continue;
	else
	  cell->ai_values[TROOP_VALUE] += 
	    (cell->connect[k]->ai_values[TROOP_VALUE] - cell->ai_values[TROOP_VALUE]) * 0.4;
    }
  */

  for(x=0;x<width;x++)
    for(y=0;y<height;y++) {
      cell = Board->cells[y][x];
      if(cell->level < 0) continue;

      /* Handle incomplete knowledge :-) */
      if(!cell->seen[side]) {
	propagate_ai_value(DECAY_NORMAL,cell,TROOP_VALUE,1.0,4);
	continue;
      }

      /* Make high elevation valuable */
      propagate_ai_value(DECAY_NORMAL,cell,CELL_VALUE,1.0*cell->level,2);

      strength = 0; owner = -1;
      for(i=0;i<Config->side_count;i++)
	if(cell->value[i] > strength) {
	  strength = cell->value[i];
	  owner = i;
	}

      if(strength == 0) {
	/* Cell is unoccupied */

	if(cell->growth >= TOWN_MIN) {
	  /* It's an empty town */
	  propagate_ai_value(DECAY_NORMAL,cell,TROOP_VALUE,50.0,15);
	  propagate_ai_value(DECAY_NORMAL,cell,TROOP_VALUE,20.0,4);
	  /* This is a d**n strict rule! */
	  cell->ai_values[TROOP_VALUE] += 1000.0;
	} else if(cell->growth > 0) {
	  /* It's an empty farm */
	  propagate_ai_value(DECAY_NORMAL,cell,TROOP_VALUE,cell->growth*1.0+1.0,4);
	  cell->ai_values[TROOP_VALUE] += cell->growth*1.0 + 1.0;
	} else {
	  /* It's an empty hex, not a farm. Make it slightly less valuable */
	  if(Config->enable_all[OPTION_DECAY])
	    cell->ai_values[TROOP_VALUE] -= 10.0 + 0.1 * Config->value_double[OPTION_DECAY][side];	    
	}
      } else if(owner == side) {
	/* AI is the dominant owner */

	/* From here on, strength represent total strength
	   of other sides in cell */
	for(i=0,strength=0;i<Config->side_count;i++)
	  if(i!=side) strength += cell->value[i];
	
	if(strength > 0) {
	  /* Cell is under attack (or we are winning an attack) */
	} else {
	  /* Cell owned by AI */

	  /* Count outgoing vectors */
	  for(j=0,outgoing=0;j<Config->direction_count;j++)
	    if(cell->dir[j]) outgoing++;
	  
	  /* Cells with many outgoing vectors have an increased troop_value
	  cell->ai_values[TROOP_VALUE] += outgoing*outgoing*outgoing * 10.0;

	  /* The cellvalue of a cell is decreased when many troops are present, this is in
	     order to evacuate crowded cells */
	  f1 = ((cell->value[side] * 1.0 / Config->max_value[side]) - 0.3) / 0.7;
	  if(f1 > 0)
	    cell->ai_values[TROOP_VALUE] -= exp(pow(f1,5.0)*15.0);

	  /* Cells where we are strong are not threated */
	  propagate_ai_value(DECAY_NORMAL,cell,THREAT,-2.0 * f1,2);

	  if(cell->growth >= TOWN_MIN) {
	    /* It's a city owned by us alone */

	    /* All cities should have a reserve of about 20% */
	    if(cell->value[side] < Config->max_value[side] * 0.2)
	      cell->ai_values[TROOP_VALUE] += 50.0 * (Config->max_value[side] * 0.2 - cell->value[side]);

	    /* Owned towns have a high value */
	    propagate_ai_value(DECAY_NORMAL,cell,CELL_VALUE,8.0,4);
	  } else if(cell->growth > 0) {
	    /* It's a farm */
	    propagate_ai_value(DECAY_NORMAL,cell,CELL_VALUE,0.5,2);
	  }
	}
      } else {
	/* Someone else is the dominant owner */

	if(cell->growth >= TOWN_MIN) {
	  /* It's an enemy city */

	  /* Enemy towns should be occupied */
	  propagate_ai_value(DECAY_NORMAL,cell,TROOP_VALUE,50.0,12);
	  propagate_ai_value(DECAY_NORMAL,cell,TROOP_VALUE,20.0,8);
	  propagate_ai_value(DECAY_NORMAL,cell,TROOP_VALUE,20.0,2);
	  cell->ai_values[TROOP_VALUE] += 20.0;

	  /* They are also threatening */
	  propagate_ai_value(DECAY_OWN_STRENGTH_INVERSE,cell,THREAT,2.0,4);
	} else {
	  /* It's a normal enemy occupied hex */

	  /* Weak enemy cells should be attacked */
	  /* hmmm... should this be measred against the AI's maxvalue or the owners? */
	  if(strength < Config->max_value[side] * 0.2)
	    cell->ai_values[TROOP_VALUE] += 20.0;

	  /* All enemy cells should be attacked, but not as important... hmm good idea? */
	  propagate_ai_value(DECAY_NORMAL,cell,TROOP_VALUE,strength*2.0/Config->max_value[side]+2.0,3);
	    
	  /* Enemies constitute a threat */
	  propagate_ai_value(DECAY_OWN_STRENGTH_INVERSE,
			     cell,THREAT,strength*2.0/Config->max_value[side],4);
	}
      }
    }

  /* Handle threats. */
  for(x=0;x<width;x++)
    for(y=0;y<height;y++) {
      cell = Board->cells[y][x];
      if(cell->level < 0) continue;

      /* Limit the threats into none or 1-10 */
      threat = cell->ai_values[THREAT];
      if(threat > 0.0) threat += 1.0;
      if(threat > 10.0) threat = 10.0;

      /* Threatened cells should increase probability of concentrating troops */
      /*      cell->ai_values[TROOP_VALUE] += cell->value[side]*5.0/Config->max_value[side] * threat; */

      /* In the intersection of valuable and threatened cells
	 the troop value should be increased */
      cell->ai_values[TROOP_VALUE] += cell->ai_values[CELL_VALUE] * threat;
    }

  /*
  printf("Cell values\n");
  for(y=0;y<height;y++) {
    for(x=0;x<width;x++) {
	cell = Board->cells[y][x];
	printf("  %2.2f",cell->ai_values[CELL_VALUE]);
      }
    printf("\n");
    }
  printf("Threat values\n");
  for(y=0;y<height;y++) {
    for(x=0;x<width;x++) {
	cell = Board->cells[y][x];
	printf("  %2.2f",cell->ai_values[THREAT]);
      }
    printf("\n");
    }
  printf("Troop values\n");
  for(y=0;y<height;y++) {
    for(x=0;x<width;x++) {
	cell = Board->cells[y][x];
	printf("  %2.2f",cell->ai_values[TROOP_VALUE]);
      }
    printf("\n");
    }
  */

  for ( ; num_actions > 0; num_actions--) {

    best_preference = 0.0;
    movement_kind = MOVEMENT_NONE; 
    X=Y=-1;
    for(x=0;x<width;x++)
      for(y=0;y<height;y++) {
	cell = Board->cells[y][x];
	if(cell->level < 0) continue;
	
	/*                                  */
	/* Consider shooting at some target */
	/*                                  */
	if(cell->ai_already_moved == 0 &&
	   Config->enable[OPTION_ARTILLERY][side] &&
	   cell->side == side) {
	  for(i=0;i<aiData[side].num_gunnery_targets;i++) {
	    range = graph_distance(cell,
				   Board->cells[aiData[side].gunnery_targets[i][1]]
				   [aiData[side].gunnery_targets[i][0]],
				   Config->enable[OPTION_ARTILLERY][side]);
	    if(range > 1 && cell->value[cell->side] > Config->value_int_all[OPTION_ARTILLERY_COST]) {
	      this_preference = (double) cell->value[side] * aiData[side].gunnery_targets[i][2] * 1.0;
	      if(this_preference > best_preference) {
		best_preference = this_preference;
		movement_kind = MOVEMENT_GUNNERY;
		X = aiData[side].gunnery_targets[i][0];
		Y = aiData[side].gunnery_targets[i][1];
		source_cell = cell;
	      }
	    }
	  }
	    
	}
	
	/*                                  */
	/* Consider normal movement actions */
	/*                                  */

	if(cell->side == side)
	  for(j=0;j<Config->direction_count;j++) {
	    cell2 = cell->connect[j];
	    if(cell2 == cell || cell2 == NULL) continue;
	    if(cell2->level < 0) continue;
	    if(cell->ai_already_moved) continue;

	    strength = 0; owner = -1;
	    for(i=0;i<Config->side_count;i++)
	      if(cell2->value[i] > strength) {
		strength = cell2->value[i];
		owner = i;
	      }	    
	    for(i=0,enemy_strength=0;i<Config->side_count;i++)
	      if(i!=side) enemy_strength += cell->value[i];
	    
	    if(cell2->ai_values[TROOP_VALUE] > cell->ai_values[TROOP_VALUE]) {
	      if(cell2->side != side && 
		 cell2->side != SIDE_NONE &&
		 cell2->side != SIDE_VOID) continue;
	      if(cell->dir[j] != 1) {
		
		/* If target is unoccupied or own cell use normal movement operations */
		this_preference = (cell2->ai_values[TROOP_VALUE] - 
				   cell->ai_values[TROOP_VALUE]) * cell->value[side];
		/*		printf("considering move from %d,%d to %d,%d with pref %f(%f)\n",
				x,y,cell2->x,cell2->y,this_preference,best_preference);*/
		if(this_preference > best_preference) {
		  /*		  printf("commiting to move from %d,%d to %d,%d with pref %f(%f)\n",
				  x,y,cell2->x,cell2->y,this_preference,best_preference); */
		  movement_kind = MOVEMENT_NORMAL_ON;
		  best_preference = this_preference;
		  source_cell = cell;
		  target_cell = cell2;
		  direction = j;		  
		} 
	      }
	    } else if(cell->dir[j] != 0) {
	      this_preference = (cell->ai_values[TROOP_VALUE] - cell2->ai_values[TROOP_VALUE]) * 
		(Config->max_value[side] - cell->value[side]) * 2.0;
	      if(this_preference > best_preference) {
		movement_kind = MOVEMENT_NORMAL_OFF;
		best_preference = this_preference;
		source_cell = cell;
		target_cell = cell2;
		direction = j;
	      }
	    }
	  }

	/*                                     */
	/* Now, consider assaults on this cell */
	/*                                     */

	strength = 0; owner = -1;
	for(i=0;i<Config->side_count;i++)
	  if(cell->value[i] > strength) {
	    strength = cell->value[i];
	    owner = i;
	  }	
	if(cell->side == side ||
	   cell->side == SIDE_NONE ||
	   cell->side == SIDE_VOID) continue;
	for(i=0,enemy_strength=0;i<Config->side_count;i++)
	  if(i!=side) enemy_strength += cell->value[i];
	if(enemy_strength == 0) continue;

	/* Add to the enemy strength for each enemy cell with incomming vectors here */
	for(j=0;j<Config->direction_count;j++) {
	  cell2 = cell->connect[j];
	  if(cell2 == cell || cell2 == NULL) continue;
	  for(k=0;k<Config->direction_count;k++)
	    if(cell2->dir[k] && cell2->connect[k] == cell) break;	      
	  if(k == Config->direction_count) break;
	  if(cell2->side != side && 
	     cell2->side != SIDE_NONE &&
	     cell2->side != SIDE_VOID &&
	     cell2->side != SIDE_FIGHT)
	    enemy_strength += 1.0 * cell2->value[cell2->side] + 0.2;
	}

	/* Calculate the odds of winning if we assult this cell, this is defined
	   as the sum of our strength in each neighbouring cell which has a preference
	   to attack this cell divided by the enemy strength in this cell. Neighbouring
	   cells which already has outgoing vectors get their strength halfed for these
	   calculations. 
	   Next, the preference of attack this cell is the sum of the cell's troop value 
	   difference of each neighbouring cell contributing to the attack (halfed for 
	   each other commitment) multiplied by the odds squared and scaled by a constant. 
	   !NOT TRUE!
	*/
	
	total_preference = 0;
	strength = 0;
	flag1 = false;
	for(j=0;j<Config->direction_count;j++) {
	  cell2 = cell->connect[j];
	  if(cell2 == cell || cell2 == NULL) continue;
	  this_strength = cell2->value[side];
	  if(this_strength <= 0) continue;
	  this_preference = cell->ai_values[TROOP_VALUE] - cell2->ai_values[TROOP_VALUE];
	  if(this_preference <= 0) continue;
	  
	  flag2 = true;	  
	  for(k=0;k<Config->direction_count;k++)
	    if(cell2->dir[k] && cell2->connect[k] == cell) flag2 == false;

	  /*	    if(cell2->dir[k]) {
	    this_strength *= 0.75;
	    this_preference *= 0.75;
	    }
	  */

	  strength += this_strength;
	  total_preference += this_preference;
	  flag1 |= flag2;
	}
	total_strength = strength;
	total_preference *= strength;
	strength = strength / enemy_strength;

	if(flag1 && strength > 1.5) {
	  total_preference *= strength * strength;
	  
	  if(total_preference > best_preference) {
	    best_preference = total_preference;
	    movement_kind = MOVEMENT_CHARGE;
	    target_cell = cell;
	    source_cell = NULL;
	    direction = -1;
	  }
	  continue;
	} 

	if(flag1) {
	  /* Put a feedback on attacking this cell, ie. increase this cells value and
	     the surrounding cells by a factor proportinal to the likely hood that
	     we could win this battle (ie. our strength/their strength) */

	  f1 = (total_preference / enemy_strength) * 0.5;
	  cell->ai_feedback_values[side][TROOP_VALUE] += f1;
	  for(j=0;j<Config->direction_count;j++) {
	    cell2 = cell->connect[j];
	    if(cell2 == cell || cell2 == NULL) continue;
	    cell2->ai_feedback_values[side][TROOP_VALUE] += f1;
	  }

	  /* Also, make this cell a potential target for gunnery attacks, if allowed, and
	     if we are not making the attack currently. */	  
	  if(strength < 1.5 &&
	     Config->enable[OPTION_ARTILLERY][side] && 
	     aiData[side].new_num_gunnery_targets < MAX_GUNNERY_TARGETS) {
	    for(j=0;j<aiData[side].new_num_gunnery_targets;j++)
	      if(x == aiData[side].new_gunnery_targets[j][0] &&
		 y == aiData[side].new_gunnery_targets[j][1]) break;
	    if(j == aiData[side].new_num_gunnery_targets) {
	      j = aiData[side].new_num_gunnery_targets++;
	      aiData[side].new_gunnery_targets[j][0] = x;
	      aiData[side].new_gunnery_targets[j][1] = y;
	      aiData[side].new_gunnery_targets[j][2] = f1;
	    }
	  }
	}

	if(cell->side == SIDE_FIGHT && strength < 1.0) {
	  /* No point in attack if we are too weak. Preference of aborting attack
	     is proportional to the number of cells attacking this cell. */
	  total_preference = 0;
	  for(j=0;j<Config->direction_count;j++) {
	    cell2 = cell->connect[j];
	    if(cell2 == cell || cell2 == NULL) continue;
	    if(cell2->value[side] == 0) continue;
	    for(k=0;k<Config->direction_count;k++)
	      if(cell2->connect[k] == cell) break;	    
	    if(cell2->dir[k])
	      total_preference += 10000.0;
	  }
	  
	  if(total_preference > best_preference) {
	    best_preference = total_preference;
	    movement_kind = MOVEMENT_ABORT_ATTACK;
	    target_cell = cell;
	    source_cell = NULL;
	    direction = -1;
	  }
	}
      }

    /*    printf("Action: %d, pref: %f\n",movement_kind,best_preference);  */
    if(movement_kind == MOVEMENT_NONE) {
      /* No action to be taken?? */
      /*      printf("no action to be taken for side %d\n",side); */
      return;
    }
    
    for(j=0;j<Config->direction_count;j++) dir[j] = 0;
    
    switch(movement_kind) {
    case MOVEMENT_NORMAL_ON:
      dir[direction]=1;
      set_move_on(source_cell,dir,0);
      source_cell->ai_already_moved = 1;
      break;
    case MOVEMENT_NORMAL_OFF:
      dir[direction]=1;
      set_move_off(source_cell,dir,0);
      source_cell->ai_already_moved = 1;
      break;
    case MOVEMENT_CHARGE:
      for(j=0;j<Config->direction_count;j++) {
	source_cell = target_cell->connect[j];
	if(source_cell == target_cell || source_cell == NULL) continue;
	if(source_cell->side != side) continue;
	if(source_cell->ai_values[TROOP_VALUE] > target_cell->ai_values[TROOP_VALUE]) continue;
	for(k=0;k<Config->direction_count;k++)
	  if(source_cell->connect[k] == target_cell) break;
	if(k==Config->direction_count)
	  printf("ERROR 42!!\n");
	dir[k]=1;
	set_move_force(source_cell,dir,0);
	dir[k]=0;
      }
      break;
    case MOVEMENT_ABORT_ATTACK:
      /*      printf("%d: %d is aborting!\n",tick,side); */
      for(j=0;j<Config->direction_count;j++) {
	source_cell = target_cell->connect[j];
	if(source_cell == target_cell || source_cell == NULL) continue;
	if(source_cell->value[side] == 0) continue;

	for(k=0;k<Config->direction_count;k++)
	  if(source_cell->connect[k] == target_cell) break;
	if(k==Config->direction_count)
	  printf("ERROR 43!!\n");
	if(!source_cell->dir[k]) continue;
	dir[k]=1;
	set_move_off(source_cell,dir,0);
	dir[k]=0;
      }
      break;
    case MOVEMENT_GUNNERY:
      /*      printf("SHOOTING from %d,%d to %d,%d\n",source_cell->x,source_cell->y,X,Y); */
      source_cell->ai_already_moved = 1;      
      run_shoot(source_cell,side,X,Y,0,TRUE);
      break;
    }
  }
}
