/* ai2.c - Pierre Bureau - 05/2001 */
/*
 * Second algorithm to handle computer players.
 *
 */


/* TODO:
 * - handle fog of war
 * - handle special options such as guns, para, etc.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

/**** x include files ****/
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#include "constant.h"
#include "extern.h"
#include "ai2.h"

#undef AI_STATUS_DEBUG
#undef AI_ATTACK_DEBUG

/* Globals */
int * routebackind;
    
/*int ** routeback;
int ** dirset;*/


/****** For the unused ScanSector ***********/
#define SECTORS 4

enum Sections {upperleft, upperright, bottomleft, bottomright};

#if SECTORS == 4
static double sectorlist[]={0,0.5,1,2,10000};
#elif SECTORS == 8
static double sectorlist[]={0,0.25,0.5,0.75,1,1.5,2,4,10000};
#else
#error Number of Sectors not defined
#endif
/**********************************************/

/* Struct for attack route algorithm */

typedef struct route_link {
    cell_type * cell;
    struct route_link * next;
}RouteLink;

/* Struct to store info for an attack */
typedef struct attack_info {
    cell_type * final_target;
    cell_type * target;
    unsigned previous; //previous resistance
    int timer;
}AttackInfo;

/*Struct to store all Ai players info*/

typedef struct player_info {
    /* Ai speed parameters */
    unsigned orga_rate;
    unsigned devel_rate;
    unsigned attack_rate;

    /* Ai skill parameters */
    float    min_att_str; //can be less than 1 for suicidal Ai
    unsigned max_perim_range;
    unsigned max_attack_range;
    float    extend_ratio;

    /* First turn flag */
    unsigned first_turn;
    
    /* Used for calculations */
    unsigned max_cell_value;

    /* Timer for victory detection */
    unsigned victorytimer;

    /* Attack info */
    unsigned max_attack; //number of authorized attack
    unsigned cur_attack; //number of current attack
    unsigned max_op_attack; //number of authorized opportunity attack
    unsigned cur_op_attack; //number of current opportunity attack
    AttackInfo attack_list[ABS_MAX_ATTACK];
    AttackInfo op_attack_list[ABS_MAX_ATTACK];
    
    /* reorganisation info */
    unsigned cur_level;
    unsigned level_size;
    cell_type * level1[MAX_TABLESIZE];
    cell_type * level2[MAX_TABLESIZE];
    cell_type ** l1;
    cell_type ** l2;

    /* Counters */
    unsigned orga_count;
    unsigned devel_count;
    unsigned attack_count;
}PlayerInfo;
 
PlayerInfo * playerlist[MAX_PLAYERS];

extern set_move_on(cell_type*,int*,int);
extern set_move_off(cell_type*,int*,int);


/********************************/
/*   Setting up the AI player   */
/********************************/

int init_algo2(int side,float skill) {
    int width,height,x,y,i;
    cell_type *cell;
    int iskill;

    iskill=skill;
    
    width = Config->board_x_size;
    height = Config->board_y_size;

    /** Init some cell's value */
    for(x=0;x<width;x++)
        for(y=0;y<height;y++) {

            cell = Board->cells[y][x];
            if(!cell) continue;

             cell->status[side]=C_DULL;
             cell->status2[side]=C_DULL;
             cell->route_tgt[side]=NULL;
             cell->need_level[side]=L_NONE;
             cell->supply/*[side]*/=0;
             cell->reserve/*[side]*/=0;
             cell->supply_scan=0;
             cell->connect_scan=0;
             cell->front_scan=0;
             cell->tmp_scan=0;
        }
    
    /** Init Player info struct **/
    playerlist[side]=(PlayerInfo*)malloc(sizeof(PlayerInfo));
    playerlist[side]->first_turn=1;
    playerlist[side]->cur_level=0;
    playerlist[side]->level_size=0;

    /** Init Ai skill data **/

    switch(iskill)
    {
    case PACIFIST :
        /* Pacifist - average development rate but hardly attacks */
        playerlist[side]->orga_rate=5;
        playerlist[side]->attack_rate=20;
        playerlist[side]->devel_rate=10;
        playerlist[side]->extend_ratio=1.5;
        playerlist[side]->max_perim_range=5;
        playerlist[side]->max_attack_range=3;
        playerlist[side]->min_att_str=1.9;
        playerlist[side]->cur_attack=0;
        playerlist[side]->max_attack=1;
        playerlist[side]->cur_op_attack=0;
        playerlist[side]->max_op_attack=1;
        break;
    case TRAINEE :
        /* Trainee - slow development, average attack skill, good organisation */
        playerlist[side]->orga_rate=1;
        playerlist[side]->attack_rate=10;
        playerlist[side]->devel_rate=15;
        playerlist[side]->extend_ratio=3;
        playerlist[side]->max_perim_range=5;
        playerlist[side]->max_attack_range=3;
        playerlist[side]->min_att_str=1.25;
        playerlist[side]->cur_attack=0;
        playerlist[side]->max_attack=2;
        playerlist[side]->cur_op_attack=0;
        playerlist[side]->max_op_attack=1;
        break;
    case COMMANDER :
        /* Commander - slow devlopment, good attack skill, good organisation */
        playerlist[side]->orga_rate=1;
        playerlist[side]->attack_rate=5;
        playerlist[side]->devel_rate=15;
        playerlist[side]->extend_ratio=3;
        playerlist[side]->max_perim_range=5;
        playerlist[side]->max_attack_range=3;
        playerlist[side]->min_att_str=1.5;
        playerlist[side]->cur_attack=0;
        playerlist[side]->max_attack=3;
        playerlist[side]->cur_op_attack=0;
        playerlist[side]->max_op_attack=2;
        break;
    case ELITE :
        /* Elite - good development, good attack skill, good organisation */
        playerlist[side]->orga_rate=1;
        playerlist[side]->attack_rate=2;
        playerlist[side]->devel_rate=10;
        playerlist[side]->extend_ratio=2;
        playerlist[side]->max_perim_range=5;
        playerlist[side]->max_attack_range=4;
        playerlist[side]->min_att_str=1.5;
        playerlist[side]->cur_attack=0;
        playerlist[side]->max_attack=3;
        playerlist[side]->cur_op_attack=0;
        playerlist[side]->max_op_attack=3;
        break;
    case QUICK :
        /* Quick - fast extension, average attack */
        playerlist[side]->orga_rate=5;
        playerlist[side]->attack_rate=10;
        playerlist[side]->devel_rate=5;
        playerlist[side]->extend_ratio=0.5;
        playerlist[side]->max_perim_range=7;
        playerlist[side]->max_attack_range=3;
        playerlist[side]->min_att_str=1.25;
        playerlist[side]->cur_attack=0;
        playerlist[side]->max_attack=2;
        playerlist[side]->cur_op_attack=0;
        playerlist[side]->max_op_attack=1;
        break;
    case PSYCO :
        /* Psyco - average development, frenetic attacks, Poor organisation */
        playerlist[side]->orga_rate=20;
        playerlist[side]->attack_rate=1;
        playerlist[side]->devel_rate=10;
        playerlist[side]->extend_ratio=1.5;
        playerlist[side]->max_perim_range=5;
        playerlist[side]->max_attack_range=3;
        playerlist[side]->min_att_str=0.5;
        playerlist[side]->cur_attack=0;
        playerlist[side]->max_attack=5;
        playerlist[side]->cur_op_attack=0;
        playerlist[side]->max_op_attack=5;
        break;
    case BARBARIAN :
        /* Barbarian - fast development, fast attack, Poor organisation */
        playerlist[side]->orga_rate=20;
        playerlist[side]->attack_rate=1;
        playerlist[side]->devel_rate=5;
        playerlist[side]->extend_ratio=1;
        playerlist[side]->max_perim_range=5;
        playerlist[side]->max_attack_range=3;
        playerlist[side]->min_att_str=1;
        playerlist[side]->cur_attack=0;
        playerlist[side]->max_attack=5;
        playerlist[side]->cur_op_attack=0;
        playerlist[side]->max_op_attack=3;
        break;
    case ULTIMATE :
        /* The Ultimate fighter */
        playerlist[side]->orga_rate=1;
        playerlist[side]->attack_rate=1;
        playerlist[side]->devel_rate=5;
        playerlist[side]->extend_ratio=1;
        playerlist[side]->max_perim_range=5;
        playerlist[side]->max_attack_range=5;
        playerlist[side]->min_att_str=1.75;
        playerlist[side]->cur_attack=0;
        playerlist[side]->max_attack=5;
        playerlist[side]->cur_op_attack=0;
        playerlist[side]->max_op_attack=5;
        break;
    default:
        /* Average settings */
        playerlist[side]->orga_rate=5;
        playerlist[side]->attack_rate=5;
        playerlist[side]->devel_rate=10;
        playerlist[side]->extend_ratio=1.5;
        playerlist[side]->max_perim_range=5;
        playerlist[side]->max_attack_range=3;
        playerlist[side]->min_att_str=1.5;
        playerlist[side]->cur_attack=0;
        playerlist[side]->max_attack=3;
        playerlist[side]->cur_op_attack=0;
        playerlist[side]->max_op_attack=2;
    }
    
    return 0;
}

/*********************************************/
/* To estimate the reserve for a given cell.
 * This function use recursive call to estimate reserve
 * for the supply network up to MAX_RESERVE_DEPTH distance.
 * When the network is modified (some routes are added or deleted),
 * reserve will change.
 */
unsigned ScanReserve(cell_type * cell,unsigned depth,int side)
{
    unsigned int i;
    cell_type * scan_cell;
    unsigned reserve=0;
    register s_char smove;
    
    if(!cell)
        return;

    smove=cell->move;

    cell->reserve/*[side]*/=0;
    
    /* If depth not too big, Scan all connected cells */
    if( depth < MAX_RESERVE_DEPTH )
        for(i=0;i<Config->direction_count;i++)
        {
            scan_cell=cell->connect[i];

            if(scan_cell != cell &&
               scan_cell->dir[routebackind[i]] &&
               scan_cell->side == side)
                reserve+=ScanReserve(scan_cell,depth+1,side);
        }

    /* Add the cell's own value */
    reserve+=cell->value[side];
    
    /* Store the cell's total reserve for future use
     * and divide the reserve by the number of leaving route
     */
    if(smove)
    {
        cell->reserve/*[side]*/=reserve/smove;
        return reserve/smove;
    }

    cell->reserve/*[side]*/=reserve;
    return reserve;
}
/*********************************************/
/* To estimate the supply for a given cell.
 * This function use recursive call to estimate supplies
 * for the complete supply network. When the network is
 * modified (some routes are added or deleted), supplies will
 * change.
 */
unsigned ScanSupply(cell_type * cell,int side)
{
    unsigned int i;
    cell_type * scan_cell;
    unsigned supply=0;
    register s_char smove;
    
    if(!cell)
        return;

    smove=cell->move;
    
    /* If the cell has been scaned before, return
       an estimated value from the previous scan.*/
    if(cell->supply_scan)
    {
      /* if(smove)
            fprintf(stderr,"(%d,%d):%u~",
                    cell->x,cell->y,cell->supply/smove);
        else
            fprintf(stderr,"(%d,%d):%u~",
                    cell->x,cell->y,cell->supply[side]);*/

        if(smove)
            return cell->supply/*[side]*//smove;
        else
            return cell->supply/*[side]*/;
    }

    cell->supply_scan=1;
    cell->supply/*[side]*/=0;
    
    /* Scan all connected cells */
    for(i=0;i<Config->direction_count;i++)
    {
        scan_cell=cell->connect[i];

        if(scan_cell != cell &&
           scan_cell->dir[routebackind[i]] &&
           scan_cell->side == side)
            supply+=ScanSupply(scan_cell,side);
    }

    /* Add the cell's own supply */
    supply+=cell->growth;
    
    /* Store the cell's total supply for future use and mark the cell as scaned*/
    cell->supply/*[side]*/=supply;

/*    if(smove)
        fprintf(stderr,"(%d,%d):%u-",cell->x,cell->y,supply/smove);
    else
        fprintf(stderr,"(%d,%d):%u-",cell->x,cell->y,supply);*/
    
    /* Divide the supply by the number of leaving route */
    if(smove)
        return supply/smove;

    return supply;
}

void ResetScanSupply(int side)
{
    unsigned x,y;
    
    for(x=0;x<Config->board_x_size;x++)
        for(y=0;y<Config->board_y_size;y++)
            Board->cells[y][x]->supply_scan=0;
}


/******************************************/
/* Allow to scan a given sector using the sectorlist.
 * Returns the number of scanned cells or 0 is an error occurs.
 * celltable must allocated and big enough to store the entire list
 * of CellCoord.
 * Sector are defined using the given x and y origins.
 */
unsigned ScanSector(cell_type **  celltable,unsigned sector,
                    double x_ref,double y_ref,enum Sections section)
{
    unsigned min_x,max_x;
    double min_y,max_y;
    unsigned i,j,ncell=0;
    unsigned topline,botline;
    double coef;

    if(sector >= SECTORS)
        return 0;
        
    switch(section)
    {
    case upperright:
        min_x=x_ref;
        max_x=Config->board_x_size;
        topline=sector+1;
        botline=sector;
        coef=1;
        break;

    case upperleft:
        min_x=0;
        max_x=x_ref;
        topline=SECTORS-sector;
        botline=topline-1;
        coef=-1;
        break;

    case bottomright:
        min_x=x_ref;
        max_x=Config->board_x_size;
        topline=SECTORS-sector;
        botline=topline-1;
        coef=-1;
        break;

    case bottomleft:
        min_x=0;
        max_x=x_ref;
        topline=sector+1;
        botline=sector;
        coef=1;
        break;
    }

    fprintf(stderr,"***minx: %u, maxx: %u, tl:%u, bl:%u\n",
            min_x,max_x,topline,botline);
    for(i=min_x;i<max_x;i++)
    {
        min_y=floor((i-x_ref)*sectorlist[botline]*coef)+y_ref;
        max_y= ceil((i-x_ref)*sectorlist[topline]*coef)+y_ref;
        
        if(min_y < 0)
            j=0;
        else
            j=min_y;
        fprintf(stderr,"miny: %f, maxy: %f ",min_y,max_y);
        while(j<max_y && j<Config->board_y_size)
        {
            celltable[ncell]=Board->cells[j][i];
            ncell++;
            fprintf(stderr,"-(%u,%u)",i,j);
            j++;
        }
        fprintf(stderr,"\n");
    }

    return ncell;
}

/******************************************/
/* Find owner of a given cell
 * Return id of the owner.or -1 if cell is unoccupied
 * Usefull for cells with fight.
 */
int FindOwner(cell_type * cell)
{
    int owner;
    unsigned strength,i;
    
    /* find cell's owner */
    strength = 0; owner = -1;
    for(i=0;i<Config->side_count;i++)
        if(cell->value[i] > strength) {
            strength = cell->value[i];
            owner = i;
        }
    return owner;
}


/*************************************************************/
/* Find and list all occupied cells.
 * celltable must be allocated and big enough to store the
 * entire list.
 * The status for the following cells is also set:
 *   - Peripheral cells (cells adjacent to free cells only)
 *   - Front cells (cells adjacent to enemy occupied cell)
 *   - Border cells (occupied cells adjacent to territory)
 *   - Free Border cells (free cells adjacent to territory)
 *
 * Returns the number of occupied cells.
 */

unsigned FindTerritoryStatus(cell_type ** territory,int side)
{
    cell_type * scan_cell;
    cell_type * cell;
    int owner;
    unsigned x,y,j,n=0;
    int isfront=0;
    int isperiph=0;
    

    for(x=0;x<Config->board_x_size;x++)
        for(y=0;y<Config->board_y_size;y++)
            if(Board->cells[y][x]->side==side)
            {
                cell=Board->cells[y][x];

                for(j=0;j<Config->direction_count;j++)
                {
                    scan_cell=cell->connect[j];

                    if(scan_cell->level < 0)
                        continue;

                    owner=scan_cell->side;

                    if(owner == SIDE_NONE)
                    {
                        scan_cell->status[side]=C_FREE_BORDER;
                        isperiph=1;
                    }
                    else
                    {
                        if(owner != side)
                        {
                            scan_cell->status[side]=C_BORDER;
                            isfront=1;
                        }
                    }
                }
        
                if(isfront)
                {
                    cell->status[side]=C_FRONT;
                    territory[n]=cell;
                    n++;
                    isfront=0;
                    isperiph=0;
                } else {
                    if(isperiph)
                    {
                        cell->status[side]=C_PERIPH;
                        territory[n]=cell;
                        n++;
                        isperiph=0;
                    }
                    else
                    {
                        cell->status[side]=C_DULL;
                        territory[n]=cell;
                        n++;
                    }
                }
            }
    return n;
}

/**************************************************************/
/* Find and list all free border cells.
 * celltable must be allocated and big enough to store the
 * entire list.
 *
 * Returns the number of free border cells.
 */
unsigned FindFreeBorder(cell_type ** celltable,int side)
{
    unsigned x,y,n=0;
    
    for(x=0;x<Config->board_x_size;x++)
        for(y=0;y<Config->board_y_size;y++)
            if(Board->cells[y][x]->status[side]==C_FREE_BORDER)
            {
                celltable[n]=Board->cells[y][x];
                n++;
            }
    return n;
}

/***********************************************/
/* Find and list all border cells (free or not)
 * celltable must be allocated and big enough to store the
 * entire list.
 *
 * Returns the number of border cells.
 */
unsigned FindBorder(cell_type ** celltable,int side)
{
    unsigned x,y,n=0;
    unsigned status;
    
    for(x=0;x<Config->board_x_size;x++)
        for(y=0;y<Config->board_y_size;y++)
        {
            status=Board->cells[y][x]->status[side];
            
            if(status == C_BORDER ||
               status == C_FREE_BORDER)
            {
                celltable[n]=Board->cells[y][x];
                n++;
            }
        }
    return n;
}
/********************************************************/
/* Find and list all peripheral cells (Front or Periph)
 *
 * Returns number of listed cells
 */
unsigned FindPeriph(cell_type ** periphlist,cell_type ** cell_list,
                    unsigned len, int side)
{
    unsigned i,n=0;
    unsigned status;
    
    for(i=0;i<len;i++)
    {
        status=cell_list[i]->status[side];

        if(status == C_FRONT ||
           status == C_PERIPH)
        {
            periphlist[n]=cell_list[i];
            n++;
        }
    }
}


/***********************************************************/
/* Find and list all front cells (adjacent to enemy)
 *
 * Returns number of listed cells
 */
unsigned FindFront(cell_type ** frontlist,cell_type ** cell_list,
                        unsigned length,int side)
{
    unsigned i,n=0;
    
    for(i=0;i<length;i++)
        if(cell_list[i]->status[side] == C_FRONT)
        {
            frontlist[n]=cell_list[i];
            n++;
        }
    
    return n;
}

/**************************************************************/
/* Reset status for the board's cells and for the given player
 *
 */
void ResetStatus(int side)
{
    unsigned x,y;
    
    for(x=0;x<Config->board_x_size;x++)
        for(y=0;y<Config->board_y_size;y++)
        {
            switch(Board->cells[y][x]->status2[side])
            {
            case C_ROUTE_TO_TGT:
                /* Reset if target is occupied */
                if(Board->cells[y][x]->route_tgt[side]->side != SIDE_NONE)
                    Board->cells[y][x]->status2[side]=C_DULL;
                break;
                
            case C_ATTACK_SUPPLY:
                /* Reset cells if attacked target is taken */
                if(Board->cells[y][x]->side != side ||
                   (Board->cells[y][x]->route_tgt[side] &&
                   Board->cells[y][x]->route_tgt[side]->status2[side] != C_TGT_UNDER_ATTACK)/*||
                   Board->cells[y][x]->status[side] != C_FRONT*/ )
                {
#ifdef AI_ATTACK_DEBUG
                    fprintf(stderr,"rem attack:(%d,%d)\n",x,y);
#endif
                    Board->cells[y][x]->route_tgt[side]=NULL;
                    Board->cells[y][x]->status2[side]=C_DULL;
                }
                break;
                
            case C_ROUTE_TO_ATTACK:
                /*Don't touch attack routes*/
                break;
            case C_INDIRECT_ATTACK:
                /*Don't touch attack target*/
                break;
            case C_DIRECT_ATTACK:
                /*Don't touch attack target*/
                break;
            case C_TGT_UNDER_ATTACK:
                /*Don't touch attack target*/
                break;
                
            default:
                if(Board->cells[y][x]->side == side)
                    Board->cells[y][x]->status2[side]=C_DULL;
            }

            Board->cells[y][x]->status[side]= C_DULL;
            Board->cells[y][x]->front_scan=0;
        }
}

void ResetAttackRoute(int side)
{
    unsigned x,y;
    
    for(x=0;x<Config->board_x_size;x++)
        for(y=0;y<Config->board_y_size;y++)
        {
            switch(Board->cells[y][x]->status2[side])
            {
            case C_ROUTE_TO_ATTACK:
                /*Reset routes*/
                Board->cells[y][x]->status2[side]=C_DULL;
                Board->cells[y][x]->route_tgt[side]=NULL;
                break;
                
            case C_INDIRECT_ATTACK:
                /*Reset all target*/
                Board->cells[y][x]->status2[side]=C_DULL;
                break;
                
            case C_DIRECT_ATTACK:
                /* Reset all target */
                Board->cells[y][x]->status2[side]=C_DULL;
                break;
            }
        }
}
/****************************************************/
/* Find the next need level, the next level includes
 * all cells, that are not part of a level, and connect to
 * the given level list
 *
 */
unsigned FindNextLevel(cell_type ** leveltab,unsigned length,
                       cell_type ** nextlevel,
                       unsigned level,int side)
{
    unsigned i,j,n=0;
    cell_type * scan_cell;
    
    for(i=0;i<length;i++)
        for(j=0;j<Config->direction_count;j++)
        {
            scan_cell=leveltab[i]->connect[j];
            if(scan_cell->side == side &&
               scan_cell->need_level[side] == L_NONE &&
               scan_cell->status2[side] != C_ROUTE_TO_TGT &&
               scan_cell->status2[side] != C_INDIRECT_TGT &&
               scan_cell->status2[side] != C_ROUTE_TO_ATTACK)
            {
                scan_cell->need_level[side]=level;
                nextlevel[n]=scan_cell;
                n++;
            }
        }
    
    return n;
}
/***************************************************/
/* Switch on routes from cells of the given level to cells of
 * lower level;
 *
 */
ConnectLevel(cell_type ** leveltab,unsigned length,unsigned level,int side)
{
    unsigned i,j;
    cell_type * scan_cell;
    int dir[MAX_DIRECTIONS]={0,0,0,0};
    
    for(i=0;i<length;i++)
    {
        for(j=0;j<Config->direction_count;j++)
        {
            scan_cell=leveltab[i]->connect[j];
            if(scan_cell->side == side &&
               scan_cell->need_level[side] == level-1)
            {
                dir[j]=1;
#ifdef ONE_CONNECT_PER_LEVEL
                break;
#endif
            }
        }
        set_move_on(leveltab[i],dir,0);
        draw_multiple_cell (leveltab[i]);

        for(j=0;j<Config->direction_count;j++)
            dir[j]=0;
    }
}

void ResetLevel(int side)
{
    unsigned x,y;
    cell_type * scan_cell;
    
    for(x=0;x<Config->board_x_size;x++)
        for(y=0;y<Config->board_y_size;y++)
        {
            scan_cell=Board->cells[y][x];

            if(scan_cell->status[side] == C_FRONT)
                scan_cell->need_level[side]=L_FRONT;
            else
                scan_cell->need_level[side]=L_NONE;
        }
}


/*******************************************************/
/* Scan board to switch off "double" routes
 *
 */
FindDoubleRoute(cell_type ** territory, unsigned t_len)
{
    unsigned i,j;
    cell_type * scan_cell;
    
    for(i=0;i<t_len;i++)
        for(j=0;j<Config->direction_count;j++)
        {
            scan_cell=territory[i]->connect[j];
            if(territory[i]->dir[j] &&
               scan_cell->dir[routebackind[j]])
            {
                //set_move_off(territory[i],dirset[j],0);
                set_move_off(territory[i],dirset+j,0);
                //set_move_off(scan_cell,routeback[j],0);
                set_move_off(scan_cell,dirset[routebackind[j]],0);
            }
        }
}


/*******************************************************/
/* Scan board to switch off useless route.
 *
 */
FindDumbRoute(cell_type ** territory, unsigned t_len,
              cell_type ** celltab,unsigned c_len,int side)
{
    unsigned x,y,j,i;
    unsigned supply_ref=0;
    unsigned supply_test=0;
    cell_type * scan_cell;

    for(i=0;i<c_len;i++)
    {
       // fprintf(stderr,"c%u:(%d,%d);",
         //       i,celltab[i]->x,celltab[i]->y);
        ScanSupply(celltab[i],side);
        //fprintf(stderr,"sup:%u",celltab[i]->supply[side]);

        supply_ref+=celltab[i]->supply/*[side]*/;
    }

   // fprintf(stderr,"ref:%u\n",supply_ref);

    for(x=0;x<t_len;x++)
    {
        scan_cell=territory[x];
        for(j=0;j<Config->direction_count;j++)
        {
            if(scan_cell->dir[j] &&
               scan_cell->connect[j]->status[side]!=C_FRONT &&
               scan_cell->connect[j]->status[side]!=C_PERIPH )
            {
                ResetScanSupply(side);
                //set_move_off(scan_cell,dirset[j],0);
                set_move_off(scan_cell,dirset[j],0);
                for(i=0;i<c_len;i++)
                {
     //               fprintf(stderr,"*");
                    ScanSupply(celltab[i],side);
       //             fprintf(stderr,"sup:%u",celltab[i]->supply[side]);
                    if(!celltab[i]->supply/*[side]*/)
                    {
                        supply_test=0;
                        break;
                    }else
                        supply_test+=celltab[i]->supply/*[side]*/;
                }

         //       fprintf(stderr,"ter(%d,%d):%u\n",
        //                scan_cell->x,scan_cell->y,supply_test);

                if(supply_test <= supply_ref)
                    //set_move_on(scan_cell,dirset[j],0);
                    set_move_on(scan_cell,dirset[j],0);
           //     else
             //       fprintf(stderr,"$(%d,%d)-%u;\n",scan_cell->x,
               //             scan_cell->y,j);

                draw_multiple_cell (scan_cell);

                supply_test=0;
            }
        }
    }
}

/***********************************************************/
void ResetScanConnect(int side)
{
    unsigned x,y;
    cell_type * cell;
    
    for(x=0;x<Config->board_x_size;x++)
        for(y=0;y<Config->board_y_size;y++)
        {
            cell=Board->cells[y][x];
            if(cell->status[side] == C_FRONT ||
               cell->status[side] == C_PERIPH )
                cell->connect_scan=0;
            else
                cell->connect_scan=C_SCAN_FLAG;
        }
}

/*************************************************************/
/* Find a route to the given cell using the connect_scan
 * level value.
 */
FindRouteTo(cell_type * cell,int side,cell_type * tgt)
{
    unsigned j;
    cell_type * scan_cell;
    
    for(j=0;j<Config->direction_count;j++)
    {
        scan_cell=cell->connect[j];
        if(scan_cell->status[side] == C_FRONT ||
           scan_cell->status[side] == C_PERIPH ||
           scan_cell->status2[side] == C_DIRECT_TGT)
            return;
        if(scan_cell->connect_scan == (cell->connect_scan-1))
        {
            if(scan_cell->status2[side] != C_ROUTE_TO_TGT)
            {
                scan_cell->status2[side]=C_ROUTE_TO_TGT;
                scan_cell->route_tgt[side]=tgt;
            }
            FindRouteTo(scan_cell,side,tgt);
            return;
        }
    }
}
/*************************************************************/
/* Find a route to the given enemy cell using the connect_scan
 * level value.
 * Route may use other enemy cells or free cells.
 *
 * Returns, total amoutn of enemy troop encountered on the route
 */

unsigned FindAttackRouteTo(cell_type * cell,RouteLink * origin,
                           unsigned length,int side)
{
    unsigned j;
    cell_type * scan_cell;
    RouteLink * scan_link;
    RouteLink newlink;
    unsigned best=~0;
    unsigned tmp;

    if(length < MAX_ROUTE_LEN)
    {
        /* Scan all cell connected with given cell */
        for(j=0;j<Config->direction_count;j++)
        {
            scan_cell=cell->connect[j];

            /* Checking for border cells with self-connection */
            if(cell == scan_cell)
                goto jump;

            /* First check if the cell has not been already visited */
            scan_link=origin;

            while(scan_link)
            {
                if(scan_cell == scan_link->cell)
                    goto jump;
                scan_link=scan_link->next;
            }

            /* If the given cell is connected to front, Stop reccursion */
            if(scan_cell->status[side] == C_FRONT ||
               scan_cell->status[side] == C_PERIPH)
            {
                cell->attack_route=scan_cell;
                return cell->value[cell->side];
            }
            else {
                /* Launch reccursion on cells with proper scan_connect */
                if(scan_cell->connect_scan != C_SCAN_FLAG &&
                   scan_cell->connect_scan >= cell->connect_scan)
                {
                    /* Update the route link with the new cell */
                    newlink.cell=cell;
                    newlink.next=origin;
                    tmp=FindAttackRouteTo(scan_cell,&newlink,length+1,side);

                    if(tmp != ~0 &&
                       tmp + cell->value[cell->side]< best)
                    {
                        best=tmp + cell->value[cell->side];
                        cell->attack_route=scan_cell;
                }
                }
            }

        jump:

        }
    }
    return best;
}
/***********************************************************/
/* After a successful FindAttackRouteTo(), this function marks
 * all the cells of the best route found as C_ROUTE_TO_ATTACK.
 * It also set the starting point for this route, cell->route_resist
 */
MarkAttackRouteTo(cell_type * cell, unsigned route_value, int side)
{
    unsigned length=0;
    cell_type * scan_cell;

    scan_cell=cell->attack_route;
    
    while( scan_cell->status[side] != C_FRONT &&
           scan_cell->status[side] != C_PERIPH &&
           length < MAX_ROUTE_LEN +5 )
    {
        scan_cell->status2[side]=C_ROUTE_TO_ATTACK;
#ifdef AI_ATTACK_DEBUG
        fprintf(stderr,"(%d,%d)->",scan_cell->x,scan_cell->y);
#endif
        scan_cell->route_resist[side]=route_value;
        scan_cell->route_tgt[side]=cell;
        scan_cell=scan_cell->attack_route;
        length++;
    }
#ifdef AI_ATTACK_DEBUG
    fprintf(stderr,"(%d,%d):%d\n",cell->x,cell->y,route_value);
#endif
}

/***************************************************/
/* Scan the enemy territory to find valuable cells such as
 * towns or weak cells.
 * The scan depth in enemy territory is MAX_ATTACK_RANGE
 */
ScanEnemy(cell_type ** borderlist,unsigned length,int side)
{
    int scan_level=-3;
    unsigned i,j,k,levlen=0;
    unsigned n=0;
    unsigned best_route;
    cell_type * level1[MAX_LEVLEN];
    cell_type * level2[MAX_LEVLEN];
    cell_type ** levptr1;
    cell_type ** levptr2;
    cell_type ** tmpptr;
    cell_type * scan_cell;
    cell_type * test_cell;

    ResetScanConnect(side);

    for(i=0;i<length;i++)
        borderlist[i]->connect_scan=-1;
    
    for(i=0;i<length;i++)
    {
        test_cell=borderlist[i];
 
        if(test_cell->growth >= TOWN_MIN &&
           test_cell->side != SIDE_NONE &&
           test_cell->status2[side] != C_TGT_UNDER_ATTACK)
            test_cell->status2[side]=C_DIRECT_ATTACK;

        for(j=0;j<Config->direction_count;j++)
        {
            scan_cell=test_cell->connect[j];
            if(scan_cell->side != side &&
               scan_cell->connect_scan == C_SCAN_FLAG &&
               scan_cell->level >= 0)
            {
                scan_cell->connect_scan=-2;
                level1[levlen]=scan_cell;
                levlen++;
            }
        }
    }

    levptr1=level1;
    levptr2=level2;

    for(k=0;k<playerlist[side]->max_attack_range;k++)
    {
        for(i=0;i<levlen;i++)
        {
            test_cell=levptr1[i];

            if(test_cell->growth >= TOWN_MIN &&
               test_cell->side != SIDE_NONE)
            {
                if(test_cell->status2[side] != C_TGT_UNDER_ATTACK)
                    test_cell->status2[side]=C_INDIRECT_ATTACK;
                best_route=FindAttackRouteTo(test_cell,NULL,0,side);
                if(best_route != ~0)
                    MarkAttackRouteTo(test_cell,best_route,side);
            }

            for(j=0;j<Config->direction_count;j++)
            {
                scan_cell=test_cell->connect[j];
                if(scan_cell->side != side &&
                   scan_cell->connect_scan == C_SCAN_FLAG &&
                   scan_cell->level >= 0)
                {
                    levptr2[n]=scan_cell;
                    scan_cell->connect_scan=scan_level;
                    n++;
                }
            }
        }
        levlen=n;
        tmpptr=levptr1;
        levptr1=levptr2;
        levptr2=tmpptr;
        scan_level--;
        n=0;
    }
}

/***********************************************/
/* Scan the closest cells around territory to find the
 * more valuable cells.
 */
ScanPerimeter(cell_type ** borderlist,unsigned length,int side)
{
    unsigned scan_level=3;
    unsigned i,j,k,levlen=0;
    unsigned n=0;
    cell_type * level1[MAX_LEVLEN];
    cell_type * level2[MAX_LEVLEN];
    cell_type ** levptr1;
    cell_type ** levptr2;
    cell_type ** tmpptr;
    cell_type * scan_cell;
    cell_type * test_cell;
    
    
    ResetScanConnect(side);

    for(i=0;i<length;i++)
        borderlist[i]->connect_scan=1;

    for(i=0;i<length;i++)
    {
        test_cell=borderlist[i];
        
        if(test_cell->growth >= TOWN_MIN)
            test_cell->status2[side]=C_DIRECT_TGT;

        for(j=0;j<Config->direction_count;j++)
        {
            scan_cell=test_cell->connect[j];
            if(scan_cell->side == SIDE_NONE &&
               scan_cell->connect_scan == C_SCAN_FLAG &&
               scan_cell->level >= 0)
            {
                scan_cell->connect_scan=2;
                level1[levlen]=scan_cell;
              levlen++;
            }
        }
    }

    levptr1=level1;
    levptr2=level2;

    for(k=0;k<playerlist[side]->max_perim_range;k++)
    {
        for(i=0;i<levlen;i++)
        {
            test_cell=levptr1[i];

            if(test_cell->growth >= TOWN_MIN)
            {
                test_cell->status2[side]=C_INDIRECT_TGT;
                FindRouteTo(test_cell,side,test_cell);
            }

            for(j=0;j<Config->direction_count;j++)
            {
                scan_cell=test_cell->connect[j];
                if(scan_cell->side == SIDE_NONE &&
                   scan_cell->connect_scan == C_SCAN_FLAG &&
                   scan_cell->level >= 0)
                {
                    levptr2[n]=scan_cell;
                    scan_cell->connect_scan=scan_level;
                    n++;
                }
            }
        }
        levlen=n;
        tmpptr=levptr1;
        levptr1=levptr2;
        levptr2=tmpptr;
        scan_level++;
        n=0;
    }
}




/*************************************/
/* Search and toggle on a route to the given
 * cell from any side controlled cell with a reserve > condition.
 *
 * Return 1 if the move is enable
 * Return 0therwise
 */
int MoveTo(cell_type * cell,int side,unsigned condition)
{
    unsigned j,i;
    cell_type * scan_cell;
    int dir[MAX_DIRECTIONS];

    for(j=0;j<Config->direction_count;j++)
    {
        for(i=0;i<Config->direction_count;i++) dir[i] = 0;
        
        scan_cell=cell->connect[j];
        if(scan_cell->side == side &&
           scan_cell->reserve > condition)
        {
            dir[routebackind[j]]=1;
            set_move_on(scan_cell,dir,0);

            return 1;
        }
    }
    
    return 0;
}

/**********************************************************/
/* Check if opportunity attack to the given cell is possible .
 * available troops must overcome the target cell value
 *
 * Returns 1 if attack is valid
 * Returns 0 if attack is not valid
 */
int CheckOpAttack(cell_type * cell,unsigned reserve_lim,int side)
{
    unsigned available_troop=0;
    unsigned j;
    cell_type * scan_cell;

    for(j=0;j<Config->direction_count;j++)
    {
        scan_cell=cell->connect[j];
        if(scan_cell->side == side &&
           scan_cell->status2[side] != C_ATTACK_SUPPLY)
        {
            /* Note: Not sure if it is better to use the reserve
             * or the cell's value only.
             */
            //available_troop+=scan_cell->reserve-reserve_lim;
            available_troop+=scan_cell->value[side]-reserve_lim;
        }
    }

#ifdef AI_ATTACK_DEBUG
    fprintf(stderr,"op(%d,%d): r=%u,t=%u\n",
            cell->x,cell->y,cell->value[cell->side],available_troop);
#endif
    
    if(available_troop > cell->value[cell->side] * playerlist[side]->min_att_str)
        return 1;

    return 0;
}

/**********************************************************/
/* Check if the given opportunity attack should continue or not.
 *
 * Return 0 if attack should continue
 * Return 1 if attack should abort
 */
int AbortOpAttack(AttackInfo * attack,unsigned reserve_lim,int side)
{
    unsigned available_troop=0;
    unsigned resist=0;
    unsigned j;
    cell_type * scan_cell;
    cell_type * target;

    target=attack->target;

    resist=target->value[target->side];

    /* Check available troops */
    for(j=0;j<Config->direction_count;j++)
    {
        scan_cell=target->connect[j];
        if(scan_cell->side == side &&
           scan_cell->dir[routebackind[j]])
        {
            /* Do we want to use the reserve here ??? */
            available_troop+=scan_cell->value[side]-reserve_lim;
        }
    }

    if(available_troop > resist * playerlist[side]->min_att_str)
        attack->timer-=10;
    else
        attack->timer+=10;

    /* Check resistance evolution */
    if(resist < attack->previous)
        attack->timer-=10;
    else
        attack->timer+=30;

    /* Increase timer anyway */
    attack->timer++;

    if(attack->timer < 0)
        attack->timer=0;

#ifdef AI_ATTACK_DEBUG
    fprintf(stderr,"abort_op(%d,%d)->(%d,%d): r=%u,t=%u,timer=%u\n",
            target->x,target->y,
            attack->final_target->x,attack->final_target->y,
            resist,available_troop,attack->timer);
#endif

    attack->previous=resist;
    
    /* Check if attack is timed out */
    if(attack->timer > MAX_ATT_TIMER)
        return 1;

    return 0;
}

/**********************************************************/
/* Launch attack to the given cell. All connected cell with
 * a reserve level over the limit will be set to attack.
 */
unsigned Attack(cell_type * cell,unsigned limit,int side)
{
    unsigned j;
    cell_type * scan_cell;
    unsigned src=0;

    for(j=0;j<Config->direction_count;j++)
    {
        scan_cell=cell->connect[j];
        if(scan_cell->side == side &&
           scan_cell->status2[side] != C_ATTACK_SUPPLY &&
           scan_cell->reserve > limit)
        {
            if(cell->status2[side] == C_TGT_UNDER_ATTACK)
                scan_cell->route_tgt[side]=cell;
            else
                scan_cell->route_tgt[side]=cell->route_tgt[side];
            
            scan_cell->status2[side]=C_ATTACK_SUPPLY;
            set_move_on(scan_cell,dirset[routebackind[j]],0);
            src++;
        }
    }

    return src;
}
/**********************************************************/
/* Disable attack to the given cell. All connected route to
 * the cell are switched off.
 */
void UnAttack(cell_type * cell,int side)
{
    unsigned j;
    cell_type * scan_cell;

    for(j=0;j<Config->direction_count;j++)
    {
        scan_cell=cell->connect[j];
        if(scan_cell->side == side)
        {
            //scan_cell->route_tgt[side]=NULL;
            
            scan_cell->status2[side]=C_DULL;
            set_move_off(scan_cell,dirset[routebackind[j]],0);
        }
    }
}
/**********************************************************/
/* Check if the given attack should continue or not.
 *
 * Return 0 if attack should continue
 * Return 1 if attack should abort
 */
int AbortAttack(AttackInfo * attack,unsigned reserve_lim,int side)
{
    unsigned available_troop=0;
    unsigned resist=0;
    unsigned j;
    cell_type * scan_cell;
    cell_type * target;

    target=attack->target;

    if(attack->final_target == target)
        resist=target->value[target->side];
    else
        resist=target->route_resist[side];

    /* Check available troops */
    for(j=0;j<Config->direction_count;j++)
    {
        scan_cell=target->connect[j];
        if(scan_cell->side == side &&
           scan_cell->dir[routebackind[j]])
        {
            available_troop+=scan_cell->reserve-reserve_lim;
        }
    }

    if(available_troop > resist * playerlist[side]->min_att_str)
        attack->timer-=10;
    else
        attack->timer+=10;

    /* Check resistance evolution */
    if(resist < attack->previous)
        attack->timer-=10;
    else
        attack->timer+=30;

    /* Increase timer anyway */
    attack->timer++;

    if(attack->timer < 0)
        attack->timer=0;

#ifdef AI_ATTACK_DEBUG
    fprintf(stderr,"abort(%d,%d)->(%d,%d): r=%u,t=%u,timer=%u\n",
            target->x,target->y,
            attack->final_target->x,attack->final_target->y,
            resist,available_troop,attack->timer);
#endif

    attack->previous=resist;
    
    /* Check if attack is timed out */
    if(attack->timer > MAX_ATT_TIMER)
        return 1;

    return 0;
}

/**********************************************************/
/* Returns the value for a given attack
 * The higer the value is, the more interesting is the cell to
 * attack
 */
float AttackValue(cell_type * cell, int side)
{
    float value;
    float resist;
    float growth;

    if(cell->status2[side] == C_DIRECT_ATTACK)
    {
        resist=cell->value[cell->side];
        growth=cell->growth;
    }
    else
    {
        resist=cell->route_resist[side];
        growth=cell->route_tgt[side]->growth;
    }

    if(resist)
        value=growth/resist;
    else
        value=MAX_ATT_VAL;
    
    return value;
}

/**********************************************************/
/* Check if attacking the given cell is possible.
 * available troops must overcome the route resistance
 *
 * Returns 1 if attack is valid
 * Returns 0 if attack is not valid
 */
int CheckAttack(cell_type * cell,unsigned reserve_lim,int side)
{
    unsigned available_troop=0;
    unsigned resist=0;
    unsigned j;
    cell_type * scan_cell;
    
    if(cell->status2[side] == C_DIRECT_ATTACK)
        resist=cell->value[cell->side];
    else
        resist=cell->route_resist[side];
    
    for(j=0;j<Config->direction_count;j++)
    {
        scan_cell=cell->connect[j];
        if(scan_cell->side == side &&
           scan_cell->status2[side] != C_ATTACK_SUPPLY)
        {
            available_troop+=scan_cell->reserve-reserve_lim;
        }
    }

#ifdef AI_ATTACK_DEBUG
    fprintf(stderr,"(%d,%d): r=%u,t=%u\n",
            cell->x,cell->y,resist,available_troop);
#endif
    
    if(available_troop > resist * playerlist[side]->min_att_str)
        return 1;

    return 0;
}


/*******************************/
/*    The main AI function     */
/*******************************/

void ai_algo2(int side) {

    cell_type * territory[MAX_TABLESIZE];
    cell_type * freeborder[MAX_TABLESIZE];
    cell_type * frontlist[MAX_TABLESIZE];
    cell_type * periphlist[MAX_TABLESIZE];
    cell_type * borderlist[MAX_TABLESIZE];
    cell_type ** tmp;

    cell_type * scan_cell;
    
    unsigned i,j;
    unsigned x,y;
    unsigned ter,freeb,front,subfront,route,border,periph;

    int tmp_calc;

    /* For the attack routine */
    cell_type * best_attack;
    float value,best_value;
    AttackInfo * scan_attack;

#ifdef WITH_HILLS_AND_FOREST
    /* For victory detector */
    int victoryflag[MAX_PLAYERS];
    int winner;
#endif

    /*
     * We need to do this init here because the direction_count
     * is not set when init_algo2 is called.
     */
    if(playerlist[side]->first_turn)
    {
#ifdef WITH_HILLS_AND_FOREST
        for(i=0;i<Config->side_count;i++)
            victoryflag[i]=0;
#endif
        playerlist[side]->victorytimer=0;
        
        playerlist[side]->max_cell_value=
            Board->shapes[side][Board->cells[0][0]->shape_index]->max_value;

        switch(Config->direction_count)
        {
        case 4 :
            /*dirset=(int**)dirset4;
             routeback=(int **)routeback4;*/
            routebackind=(int *)routebackind4;
            break;
        case 6 :
            /*dirset=(int **)dirset6;
             routeback=(int **)routeback6;*/
            routebackind=(int *)routebackind6;
            break;
        default:
            printf("Error - Ai algo 2 doesn't support this tile\n");
            exit(0);
        }
    }

    /* Find the territory for this player */
    ter=FindTerritoryStatus(territory,side);
    
    /* Find Free border cells and border cells */
    freeb=FindFreeBorder(freeborder,side);
    border=FindBorder(borderlist,side);
    front=FindFront(frontlist,territory,ter,side);
    periph=FindPeriph(periphlist,territory,ter,side);


    /* On first turn always scan perimeter and Reserve */
    if(playerlist[side]->first_turn)
    {
        for(i=0;i<ter;i++)
            ScanReserve(territory[i],0,side);
        ScanPerimeter(freeborder,freeb,side);
        playerlist[side]->first_turn=0;
    }

    /** Scan supply and reserve for each front cell */
    for(i=0;i<periph;i++)
    {
        ScanSupply(periphlist[i],side);
        ScanReserve(periphlist[i],0,side);
    }

    /** Run the Reorganisation routine  - can take several cycles **/
    if( playerlist[side]->orga_count > playerlist[side]->orga_rate)
    {

        /*
         *  TODO: manage a better defensive system, aborting a route
         *        when attacked should use a debounce algorithm with a
         *        timer similar to the attack management system.
         */

        /* Remove routes in front cells except Attack supplies */
        for(i=0;i<front;i++)
            if( frontlist[i]->status[side] == C_FRONT &&
                frontlist[i]->status2[side] != C_ATTACK_SUPPLY &&
                frontlist[i]->status2[side] != C_ROUTE_TO_TGT )
                for(j=0;j<Config->direction_count;j++)
                    if(frontlist[i]->connect[j]->side != SIDE_FIGHT)
                    {
                        set_move_off(frontlist[i],dirset[j],0);
                        draw_multiple_cell (frontlist[i]);
                    }

        if(playerlist[side]->level_size)
        {
            playerlist[side]->level_size=
                FindNextLevel(playerlist[side]->l1,playerlist[side]->level_size,
                              playerlist[side]->l2,playerlist[side]->cur_level,side);
            if(!(playerlist[side]->level_size))
            {
                playerlist[side]->orga_count=0;
                /* Try to remove useless route when all levels are connected */
                FindDoubleRoute(territory,ter);
              //FindDumbRoute(territory,ter,enemylist,enemy,side);
            }

            ConnectLevel(playerlist[side]->l2,playerlist[side]->level_size,
                         playerlist[side]->cur_level,side);
            tmp=playerlist[side]->l1;
            playerlist[side]->l1=playerlist[side]->l2;
            playerlist[side]->l2=tmp;
            playerlist[side]->cur_level++;
        }
        else
        {
            ResetLevel(side);
            
            /* Connect the first level */
            playerlist[side]->level_size=
                FindNextLevel(frontlist,front,playerlist[side]->level1,2,side);
            ConnectLevel(playerlist[side]->level1,
                         playerlist[side]->level_size,2,side);
            playerlist[side]->l1=playerlist[side]->level1;
            playerlist[side]->l2=playerlist[side]->level2;;
            playerlist[side]->cur_level=3;
        }
    }

    /** Run the development routine **/
    if(freeb && playerlist[side]->devel_count > playerlist[side]->devel_rate)
    {
        /* Compute the minimum reserve needed to extend */
        tmp_calc=
            playerlist[side]->max_cell_value*playerlist[side]->extend_ratio;
        
        /* Look for valuable free cells */
        ScanPerimeter(freeborder,freeb,side);
        /* Move to valuable free cells*/
        for(i=0;i<freeb;i++)
        {
            if(freeborder[i]->status2[side] == C_DIRECT_TGT ||
               freeborder[i]->status2[side] == C_ROUTE_TO_TGT ||
               freeborder[i]->status2[side] == C_INDIRECT_TGT
              )
                MoveTo(freeborder[i],side,0);
            else
                MoveTo(freeborder[i],side,tmp_calc);

        }
        playerlist[side]->devel_count=0;
    }

    /** Run the attack routine **/
    if(border &&
       playerlist[side]->attack_count > playerlist[side]->attack_rate)
    {
        /*
         * TODO:
         *   - Use a better way of detecting if a new ScanEnemy is needed
         */
        
        /* If needed reset Attack route status and Find new routes */
        if(playerlist[side]->cur_attack == 0)
        {
            ResetAttackRoute(side);
            ScanEnemy(borderlist,border,side);
        }

        /* Update number of attack */

        /*
         * TODO:
         *  -use a dynamic number of attack (depending on global situation)
         *  -use a dynamic minimum reserve level to attack
         */

        /* Check if player can launch new attack */
        if(playerlist[side]->cur_attack < playerlist[side]->max_attack)
        {
            best_value=0;
            best_attack=NULL;

            /* Look for the most valuable route to attack
             * aiming at Target not already under attack
             */
            for(i=0;i<border;i++)
            {
                scan_cell=borderlist[i];
                /* Look at Direct target */
                if(scan_cell->status2[side] == C_DIRECT_ATTACK &&
                   CheckAttack(scan_cell,0,side))
                {
                    value=AttackValue(scan_cell,side);
                    if(value > best_value)
                    {
                        best_value=value;
                        best_attack=scan_cell;
                    }
                }
                /* Look at distant target */
                if(scan_cell->status2[side] == C_ROUTE_TO_ATTACK &&
                   scan_cell->route_tgt[side]->status2[side] == C_INDIRECT_ATTACK &&
                   CheckAttack(scan_cell,0,side))
                {
                    value=AttackValue(scan_cell,side);
                    if(value > best_value)
                    {
                        best_value=value;
                        best_attack=scan_cell;
                    }
                }
            }
            /* Launch the most valuable attack */
            if(best_attack)
            {
                scan_attack=&(playerlist[side]->attack_list[playerlist[side]->cur_attack]);

                scan_attack->target=best_attack;
                scan_attack->timer=0;
                scan_attack->previous=~0;
                /* Mark target as under attack */
                if(best_attack->status2[side] == C_DIRECT_ATTACK)
                {
                    best_attack->status2[side]=C_TGT_UNDER_ATTACK;
                    scan_attack->final_target=best_attack;
                }
                else
                {
                    best_attack->route_tgt[side]->status2[side]=C_TGT_UNDER_ATTACK;
                    scan_attack->final_target=best_attack->route_tgt[side];
                }
                Attack(best_attack,0,side);

                playerlist[side]->cur_attack++;
            }
        }
        
        /* Check if player can launch new opportunity attack */
        if(playerlist[side]->cur_op_attack < playerlist[side]->max_op_attack)
        {
            for(i=0;i<border;i++)
            {
                scan_cell=borderlist[i];
                
                if(playerlist[side]->cur_op_attack < playerlist[side]->max_op_attack &&
                   scan_cell->status2[side] != C_TGT_UNDER_ATTACK &&
                   scan_cell->status[side] == C_BORDER &&
                   CheckOpAttack(scan_cell,0,side))
                {
                    scan_attack=&(playerlist[side]->op_attack_list[playerlist[side]->cur_op_attack]);

                    scan_attack->target=scan_cell;
                    scan_attack->timer=0;
                    scan_attack->previous=~0;
                    scan_attack->final_target=scan_cell;

                    scan_cell->status2[side]=C_TGT_UNDER_ATTACK;

                    Attack(scan_cell,0,side);
                    playerlist[side]->cur_op_attack++;
                }
            }
        }

        playerlist[side]->attack_count=0;
    }

    /* Monitor attacks evolution */
    for(i=0;i<playerlist[side]->cur_attack;i++)
    {
        /* Note: Each time an attack is canceled or completed,
         *       the last attack in the list is moved and
         *       skipped. We could work arround that problem.
         *       Later...
         */

        scan_attack=&(playerlist[side]->attack_list[i]);

        /* Abort attack with little chance of success */
        if(AbortAttack(scan_attack,0,side))
        {

#ifdef AI_ATTACK_DEBUG
            fprintf(stderr,"abort\n");
#endif
            UnAttack(scan_attack->target,side);
            scan_attack->final_target->status2[side]=C_DULL;

            /* Copy last attack in place of the canceled one */
            tmp_calc=playerlist[side]->cur_attack-1;

            if(tmp_calc != i)
                memcpy(scan_attack,
                       &(playerlist[side]->attack_list[tmp_calc]),
                       sizeof(AttackInfo));

            playerlist[side]->cur_attack--;

        } else {
            /* Change attack route if a better route is found */

            /*
             * TODO
             */

            /* Maintain successful attacks */
            if(scan_attack->target != scan_attack->final_target &&
               scan_attack->target->side == side)
            {
                /* Look for next cell to attack */
                for(j=0;j<Config->direction_count;j++)
                {
                    scan_cell=scan_attack->target->connect[j];
                    if(scan_cell == scan_attack->final_target ||
                       (scan_cell->status2[side] == C_ROUTE_TO_ATTACK &&
                        scan_cell->route_tgt[side] == scan_attack->final_target))
                    {
                        /* update attack & suspend attack if reserves are low */
                        if(Attack(scan_cell,0,side))
                        {
                            /* Reset the abort timer */
                            scan_attack->timer=0;
                            scan_attack->target=scan_cell;
                        }
#ifdef AI_ATTACK_DEBUG
                        else
                            fprintf(stderr,"suspend\n");
#endif
                        /*
                         * DO we need the following break ???
                         */
                        //break;
                    }
                }
            }

            /* Clear completed attack */
            if(scan_attack->final_target->side == side)
            {
                scan_attack->final_target->status2[side]=C_DULL;

                /* Copy last attack in place of the completed one */
                tmp_calc=playerlist[side]->cur_attack-1;
                if(tmp_calc != i)
                    memcpy(scan_attack,
                           &(playerlist[side]->attack_list[tmp_calc]),
                           sizeof(AttackInfo));

                playerlist[side]->cur_attack--;
            }
        }
    }

    /* Monitor opportunity attacks evolution */
    for(i=0;i<playerlist[side]->cur_op_attack;i++)
    {
        scan_attack=&(playerlist[side]->op_attack_list[i]);

        /* Abort attack with little chance of success */
        if(AbortOpAttack(scan_attack,0,side))
        {
            UnAttack(scan_attack->target,side);
            scan_attack->final_target->status2[side]=C_DULL;

            /* Copy last attack in place of the canceled one */
            tmp_calc=playerlist[side]->cur_op_attack-1;

            if(tmp_calc != i)
                memcpy(scan_attack,
                       &(playerlist[side]->op_attack_list[tmp_calc]),
                       sizeof(AttackInfo));

            playerlist[side]->cur_op_attack--;

        } else {
            /* Clear completed attack */
            if(scan_attack->final_target->side == side)
            {
                scan_attack->final_target->status2[side]=C_DULL;

                /* Copy last attack in place of the completed one */
                tmp_calc=playerlist[side]->cur_op_attack-1;
                if(tmp_calc != i)
                    memcpy(scan_attack,
                           &(playerlist[side]->op_attack_list[tmp_calc]),
                           sizeof(AttackInfo));

                playerlist[side]->cur_op_attack--;
            }
        }
    }

#ifdef WITH_VICTORY_DETECTOR
    /*
     * TODO: more complex victories (alies, objectives)
     */

    /* Check if a side has won the game */

    for(x=0;x<Config->board_x_size;x++)
        for(y=0;y<Config->board_y_size;y++)
        {
            tmp_calc=Board->cells[y][x]->side;

            if(tmp_calc != SIDE_NONE &&
               tmp_calc != SIDE_FIGHT)
                if(Board->cells[y][x]->value[tmp_calc] > 1)
                    victoryflag[tmp_calc]=1;
        }

    tmp_calc=0;

    for(i=0;i<Config->side_count;i++)
        if(victoryflag[i])
        {
            tmp_calc++;
            victoryflag[i]=0;
            winner=i;
        }

    if(tmp_calc == 1)
        playerlist[side]->victorytimer++;

    if(playerlist[side]->victorytimer > MAX_VICTORY_TIMER)
        exit(winner+10);
#endif

    ResetStatus(side);
    ResetScanSupply(side);

    playerlist[side]->orga_count++;
    playerlist[side]->devel_count++;
    playerlist[side]->attack_count++;


#ifdef AI_STATUS_DEBUG
    for(y=0;y<Config->board_y_size;y++)
    {
        for(x=0;x<Config->board_x_size;x++)
        {
            if(Board->cells[y][x]->status2[side]==C_DULL)
                fprintf(stderr,"%u.",Board->cells[y][x]->status[side]);
            else
                fprintf(stderr,"%u.",Board->cells[y][x]->status2[side]);
            //fprintf(stderr,"%u.",Board->cells[y][x]->need_level[side]);
        }

        fprintf(stderr," ----- ");
        
        for(x=0;x<Config->board_x_size;x++)
        {
         /*   if(Board->cells[y][x]->status2[side]==C_DULL)
                fprintf(stderr,"%u.",Board->cells[y][x]->status[side]);
            else
                fprintf(stderr,"%u.",Board->cells[y][x]->status2[side]);*/
            fprintf(stderr,"%u.",Board->cells[y][x]->side);
        }
        fprintf(stderr,"\n");
    }
#endif

#if (defined(AI_STATUS_DEBUG) || defined(AI_ATTACK_DEBUG))
    fprintf(stderr,"****\n");
#endif

}
